namespace K3D
{
   using System;
   using System.Drawing;
   using System.Collections;
   using System.ComponentModel;
   using System.Windows.Forms;

   using K3D.Geometry;
   using K3D.Services;

   using Tao.OpenGl;


	/// <summary>
	/// Light Controls Dialog.
	/// </summary>
	public sealed class LightsDialog : System.Windows.Forms.Form
	{
      // ------------------------------------------------------------------------------
      // Construction

      /// <summary>
      /// Constructor
      /// </summary>
		public LightsDialog()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
         
         // set the active light checkboxes
         LightInfo info = new LightInfo();
         info.DeserialiseFromRegistry(REG_LIGHT1);
         this.checkLight1.Checked = info.Active;
         info.DeserialiseFromRegistry(REG_LIGHT2);
         this.checkLight2.Checked = info.Active;
         info.DeserialiseFromRegistry(REG_LIGHT3);
         this.checkLight3.Checked = info.Active;
         info.DeserialiseFromRegistry(REG_LIGHT4);
         this.checkLight4.Checked = info.Active;
         
         // bind control to the current light info
         BindCurrentLightDetails();
         
         // add event handlers
         this.checkLight1.CheckedChanged += new System.EventHandler(this.OnActiveLightCheckChanged);
         this.checkLight2.CheckedChanged += new System.EventHandler(this.OnActiveLightCheckChanged);
         this.checkLight3.CheckedChanged += new System.EventHandler(this.OnActiveLightCheckChanged);
         this.checkLight4.CheckedChanged += new System.EventHandler(this.OnActiveLightCheckChanged);
      }


      // ------------------------------------------------------------------------------
      // Public API

      /// <summary>
      /// Refresh the specified OpenGL light with it's new settings
      /// </summary>
      /// <param name="lightIndex">GL_LIGHTX to setup</param>
      /// <param name="info">LightInfo structure to set data from</param>
      public static void UpdateGLLight(int lightIndex, LightInfo info)
      {
         Gl.glLightfv(lightIndex, Gl.GL_AMBIENT, LightAmbient);
         float[] diffuse = new float[4] {info.Colour.Red * info.Intensity, info.Colour.Green * info.Intensity, info.Colour.Blue * info.Intensity, 1.0f};
         Gl.glLightfv(lightIndex, Gl.GL_DIFFUSE, diffuse);
         float[] specular = new float[4] {info.Specular, info.Specular, info.Specular, 1.0f};
         Gl.glLightfv(lightIndex, Gl.GL_SPECULAR, specular);
         float[] position = new float[4] {-info.Position.X, info.Position.Y, info.Position.Z, 1.0f};
         Gl.glLightfv(lightIndex, Gl.GL_POSITION, position);
         if (info.Active == true)
         {
            Gl.glEnable(lightIndex);
         }
         else
         {
            Gl.glDisable(lightIndex);
         }
      }


      // ------------------------------------------------------------------------------
      // Private methods and event handlers

      /// <summary>
      /// Currently selected light changed
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
      private void OnLightChanged(object sender, System.EventArgs e)
      {
         // de-serialise the select light data from the registry and show it
         RadioButton button = sender as RadioButton;
         int newIndex = (button.TabIndex);
         if (newIndex != m_currentLightIndex)
         {
            m_currentLightIndex = newIndex;
            BindCurrentLightDetails();
         }
      }

      /// <summary>
      /// User is enabling/disabling specific lights
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
      private void OnActiveLightCheckChanged(object sender, System.EventArgs e)
      {
         CheckBox checkbox = sender as CheckBox;
         
         int lightIndex = (checkbox.TabIndex);
         
         LightInfo info = new LightInfo();
         info.DeserialiseFromRegistry(REG_LIGHT + lightIndex);
         info.Active = checkbox.Checked;
         info.SerialiseToRegistry(REG_LIGHT + lightIndex);
         
         UpdateGLLight(Gl.GL_LIGHT0 + lightIndex, info);
         ((K3DGLForm)this.Owner).Renderer.Invalidate();
      }

      /// <summary>
      /// Using the current light info, bind the controls to display the details
      /// </summary>
      private void BindCurrentLightDetails()
      {
         LightInfo info = GetCurrentLightInfo();
         
         // disable events for bound controls
			this.hScrollBarXPos.ValueChanged -= new System.EventHandler(this.OnXPositionChanged);
			this.hScrollBarYPos.ValueChanged -= new System.EventHandler(this.OnYPositionChanged);
			this.hScrollBarZPos.ValueChanged -= new System.EventHandler(this.OnZPositionChanged);
			
			this.trackSpecular.ValueChanged -= new System.EventHandler(this.OnSpecularIntensityChanged);
			this.trackIntensity.ValueChanged -= new System.EventHandler(this.OnIntensityValueChanged);
			
			// bind control values
         this.trackIntensity.Value = (int)(info.Intensity * 100.0f);
         this.trackSpecular.Value = (int)(info.Specular * 100.0f);
         this.btnColour.BackColor = info.Colour.ToColor();
         
         this.hScrollBarXPos.Value = (int)info.Position.X;
         this.hScrollBarYPos.Value = (int)info.Position.Y;
         this.hScrollBarZPos.Value = (int)info.Position.Z;
			
         // enable events for bound controls
			this.hScrollBarXPos.ValueChanged += new System.EventHandler(this.OnXPositionChanged);
			this.hScrollBarYPos.ValueChanged += new System.EventHandler(this.OnYPositionChanged);
			this.hScrollBarZPos.ValueChanged += new System.EventHandler(this.OnZPositionChanged);
			
			this.trackSpecular.ValueChanged += new System.EventHandler(this.OnSpecularIntensityChanged);
			this.trackIntensity.ValueChanged += new System.EventHandler(this.OnIntensityValueChanged);
		}

      /// <summary>
      /// Get the LightInfo structure for the currently selected light
      /// </summary>
      /// <returns></returns>
      private LightInfo GetCurrentLightInfo()
      {
         LightInfo info = new LightInfo();
         info.DeserialiseFromRegistry(REG_LIGHT + m_currentLightIndex);
         
         return info;
      }

		/// <summary>
		/// Update the currently selected light with the specified LightInfo
		/// </summary>
		/// <param name="info"></param>
		private void UpdateCurrentLight(LightInfo info)
		{
			UpdateGLLight(Gl.GL_LIGHT0 + m_currentLightIndex, info);
			((K3DGLForm)this.Owner).Renderer.Invalidate();
			info.SerialiseToRegistry(REG_LIGHT + m_currentLightIndex);
		}

		/// <summary>
		/// Colour button clicked
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void OnColourButtonClick(object sender, System.EventArgs e)
		{
			this.colorDialog.Color = this.btnColour.BackColor;
			if (this.colorDialog.ShowDialog() == DialogResult.OK)
			{
				this.btnColour.BackColor = this.colorDialog.Color;
				LightInfo lightInfo = GetCurrentLightInfo();
				lightInfo.Colour = new ColourTriple(this.colorDialog.Color);
				UpdateCurrentLight(lightInfo);
			}
		}

		private void OnIntensityValueChanged(object sender, System.EventArgs e)
		{
			TrackBar tracker = (TrackBar)sender;
			LightInfo lightInfo = GetCurrentLightInfo();
			lightInfo.Intensity = ((float)tracker.Value) / 100.0f;
			UpdateCurrentLight(lightInfo);
		}

		private void OnSpecularIntensityChanged(object sender, System.EventArgs e)
		{
			TrackBar tracker = (TrackBar)sender;
			LightInfo lightInfo = GetCurrentLightInfo();
			lightInfo.Specular = ((float)tracker.Value) / 100.0f;
			UpdateCurrentLight(lightInfo);
		}

		private void OnXPositionChanged(object sender, System.EventArgs e)
		{
			ScrollBar scrollbar = (ScrollBar)sender;
			LightInfo lightInfo = GetCurrentLightInfo();
			lightInfo.Position.X = (float)scrollbar.Value;
			UpdateCurrentLight(lightInfo);
		}

		private void OnYPositionChanged(object sender, System.EventArgs e)
		{
			ScrollBar scrollbar = (ScrollBar)sender;
			LightInfo lightInfo = GetCurrentLightInfo();
			lightInfo.Position.Y = (float)scrollbar.Value;
			UpdateCurrentLight(lightInfo);
		}

		private void OnZPositionChanged(object sender, System.EventArgs e)
		{
			ScrollBar scrollbar = (ScrollBar)sender;
			LightInfo lightInfo = GetCurrentLightInfo();
			lightInfo.Position.Z = (float)scrollbar.Value;
			UpdateCurrentLight(lightInfo);
		}


      // ------------------------------------------------------------------------------
      // Designer code

		#region Windows Form Designer generated code
      /// <summary>
      /// Clean up any resources being used.
      /// </summary>
      protected override void Dispose( bool disposing )
      {
         if( disposing )
         {
            if(components != null)
            {
               components.Dispose();
            }
         }
         base.Dispose( disposing );
      }
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(LightsDialog));
			this.groupBox1 = new System.Windows.Forms.GroupBox();
			this.radioLight4 = new System.Windows.Forms.RadioButton();
			this.radioLight3 = new System.Windows.Forms.RadioButton();
			this.radioLight2 = new System.Windows.Forms.RadioButton();
			this.radioLight1 = new System.Windows.Forms.RadioButton();
			this.groupBox2 = new System.Windows.Forms.GroupBox();
			this.checkLight4 = new System.Windows.Forms.CheckBox();
			this.checkLight3 = new System.Windows.Forms.CheckBox();
			this.checkLight2 = new System.Windows.Forms.CheckBox();
			this.checkLight1 = new System.Windows.Forms.CheckBox();
			this.groupBox3 = new System.Windows.Forms.GroupBox();
			this.hScrollBarZPos = new System.Windows.Forms.HScrollBar();
			this.hScrollBarYPos = new System.Windows.Forms.HScrollBar();
			this.hScrollBarXPos = new System.Windows.Forms.HScrollBar();
			this.trackSpecular = new System.Windows.Forms.TrackBar();
			this.label1 = new System.Windows.Forms.Label();
			this.trackIntensity = new System.Windows.Forms.TrackBar();
			this.lblIntensity = new System.Windows.Forms.Label();
			this.btnColour = new System.Windows.Forms.Button();
			this.lblColour = new System.Windows.Forms.Label();
			this.label2 = new System.Windows.Forms.Label();
			this.label3 = new System.Windows.Forms.Label();
			this.label4 = new System.Windows.Forms.Label();
			this.groupBox4 = new System.Windows.Forms.GroupBox();
			this.btnClose = new System.Windows.Forms.Button();
			this.colorDialog = new System.Windows.Forms.ColorDialog();
			this.groupBox1.SuspendLayout();
			this.groupBox2.SuspendLayout();
			this.groupBox3.SuspendLayout();
			((System.ComponentModel.ISupportInitialize)(this.trackSpecular)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.trackIntensity)).BeginInit();
			this.groupBox4.SuspendLayout();
			this.SuspendLayout();
			// 
			// groupBox1
			// 
			this.groupBox1.Controls.Add(this.radioLight4);
			this.groupBox1.Controls.Add(this.radioLight3);
			this.groupBox1.Controls.Add(this.radioLight2);
			this.groupBox1.Controls.Add(this.radioLight1);
			this.groupBox1.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.groupBox1.Location = new System.Drawing.Point(8, 8);
			this.groupBox1.Name = "groupBox1";
			this.groupBox1.Size = new System.Drawing.Size(80, 120);
			this.groupBox1.TabIndex = 0;
			this.groupBox1.TabStop = false;
			this.groupBox1.Text = "Current Light";
			// 
			// radioLight4
			// 
			this.radioLight4.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.radioLight4.Location = new System.Drawing.Point(12, 88);
			this.radioLight4.Name = "radioLight4";
			this.radioLight4.Size = new System.Drawing.Size(56, 24);
			this.radioLight4.TabIndex = 3;
			this.radioLight4.Text = "Light 4";
			this.radioLight4.CheckedChanged += new System.EventHandler(this.OnLightChanged);
			// 
			// radioLight3
			// 
			this.radioLight3.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.radioLight3.Location = new System.Drawing.Point(12, 64);
			this.radioLight3.Name = "radioLight3";
			this.radioLight3.Size = new System.Drawing.Size(56, 24);
			this.radioLight3.TabIndex = 2;
			this.radioLight3.Text = "Light 3";
			this.radioLight3.CheckedChanged += new System.EventHandler(this.OnLightChanged);
			// 
			// radioLight2
			// 
			this.radioLight2.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.radioLight2.Location = new System.Drawing.Point(12, 40);
			this.radioLight2.Name = "radioLight2";
			this.radioLight2.Size = new System.Drawing.Size(56, 24);
			this.radioLight2.TabIndex = 1;
			this.radioLight2.Text = "Light 2";
			this.radioLight2.CheckedChanged += new System.EventHandler(this.OnLightChanged);
			// 
			// radioLight1
			// 
			this.radioLight1.Checked = true;
			this.radioLight1.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.radioLight1.Location = new System.Drawing.Point(12, 16);
			this.radioLight1.Name = "radioLight1";
			this.radioLight1.Size = new System.Drawing.Size(56, 24);
			this.radioLight1.TabIndex = 0;
			this.radioLight1.TabStop = true;
			this.radioLight1.Text = "Light 1";
			this.radioLight1.CheckedChanged += new System.EventHandler(this.OnLightChanged);
			// 
			// groupBox2
			// 
			this.groupBox2.Controls.Add(this.checkLight4);
			this.groupBox2.Controls.Add(this.checkLight3);
			this.groupBox2.Controls.Add(this.checkLight2);
			this.groupBox2.Controls.Add(this.checkLight1);
			this.groupBox2.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.groupBox2.Location = new System.Drawing.Point(8, 136);
			this.groupBox2.Name = "groupBox2";
			this.groupBox2.Size = new System.Drawing.Size(80, 120);
			this.groupBox2.TabIndex = 1;
			this.groupBox2.TabStop = false;
			this.groupBox2.Text = "Active Lights";
			// 
			// checkLight4
			// 
			this.checkLight4.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.checkLight4.Location = new System.Drawing.Point(12, 88);
			this.checkLight4.Name = "checkLight4";
			this.checkLight4.Size = new System.Drawing.Size(60, 24);
			this.checkLight4.TabIndex = 3;
			this.checkLight4.Text = "Light 4";
			// 
			// checkLight3
			// 
			this.checkLight3.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.checkLight3.Location = new System.Drawing.Point(12, 64);
			this.checkLight3.Name = "checkLight3";
			this.checkLight3.Size = new System.Drawing.Size(60, 24);
			this.checkLight3.TabIndex = 2;
			this.checkLight3.Text = "Light 3";
			// 
			// checkLight2
			// 
			this.checkLight2.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.checkLight2.Location = new System.Drawing.Point(12, 40);
			this.checkLight2.Name = "checkLight2";
			this.checkLight2.Size = new System.Drawing.Size(60, 24);
			this.checkLight2.TabIndex = 1;
			this.checkLight2.Text = "Light 2";
			// 
			// checkLight1
			// 
			this.checkLight1.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.checkLight1.Location = new System.Drawing.Point(12, 16);
			this.checkLight1.Name = "checkLight1";
			this.checkLight1.Size = new System.Drawing.Size(60, 24);
			this.checkLight1.TabIndex = 0;
			this.checkLight1.Text = "Light 1";
			// 
			// groupBox3
			// 
			this.groupBox3.Controls.Add(this.hScrollBarZPos);
			this.groupBox3.Controls.Add(this.hScrollBarYPos);
			this.groupBox3.Controls.Add(this.hScrollBarXPos);
			this.groupBox3.Controls.Add(this.trackSpecular);
			this.groupBox3.Controls.Add(this.label1);
			this.groupBox3.Controls.Add(this.trackIntensity);
			this.groupBox3.Controls.Add(this.lblIntensity);
			this.groupBox3.Controls.Add(this.btnColour);
			this.groupBox3.Controls.Add(this.lblColour);
			this.groupBox3.Controls.Add(this.label2);
			this.groupBox3.Controls.Add(this.label3);
			this.groupBox3.Controls.Add(this.label4);
			this.groupBox3.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.groupBox3.Location = new System.Drawing.Point(96, 8);
			this.groupBox3.Name = "groupBox3";
			this.groupBox3.Size = new System.Drawing.Size(216, 248);
			this.groupBox3.TabIndex = 2;
			this.groupBox3.TabStop = false;
			this.groupBox3.Text = "Current Light Settings";
			// 
			// hScrollBarZPos
			// 
			this.hScrollBarZPos.Location = new System.Drawing.Point(16, 216);
			this.hScrollBarZPos.Maximum = 250;
			this.hScrollBarZPos.Minimum = -250;
			this.hScrollBarZPos.Name = "hScrollBarZPos";
			this.hScrollBarZPos.Size = new System.Drawing.Size(184, 15);
			this.hScrollBarZPos.TabIndex = 5;
			this.hScrollBarZPos.ValueChanged += new System.EventHandler(this.OnZPositionChanged);
			// 
			// hScrollBarYPos
			// 
			this.hScrollBarYPos.Location = new System.Drawing.Point(16, 184);
			this.hScrollBarYPos.Maximum = 250;
			this.hScrollBarYPos.Minimum = -250;
			this.hScrollBarYPos.Name = "hScrollBarYPos";
			this.hScrollBarYPos.Size = new System.Drawing.Size(184, 15);
			this.hScrollBarYPos.TabIndex = 4;
			this.hScrollBarYPos.ValueChanged += new System.EventHandler(this.OnYPositionChanged);
			// 
			// hScrollBarXPos
			// 
			this.hScrollBarXPos.Location = new System.Drawing.Point(16, 152);
			this.hScrollBarXPos.Maximum = 250;
			this.hScrollBarXPos.Minimum = -250;
			this.hScrollBarXPos.Name = "hScrollBarXPos";
			this.hScrollBarXPos.Size = new System.Drawing.Size(184, 15);
			this.hScrollBarXPos.TabIndex = 3;
			this.hScrollBarXPos.ValueChanged += new System.EventHandler(this.OnXPositionChanged);
			// 
			// trackSpecular
			// 
			this.trackSpecular.LargeChange = 20;
			this.trackSpecular.Location = new System.Drawing.Point(88, 96);
			this.trackSpecular.Maximum = 100;
			this.trackSpecular.Name = "trackSpecular";
			this.trackSpecular.Size = new System.Drawing.Size(120, 45);
			this.trackSpecular.SmallChange = 10;
			this.trackSpecular.TabIndex = 2;
			this.trackSpecular.TickFrequency = 10;
			this.trackSpecular.ValueChanged += new System.EventHandler(this.OnSpecularIntensityChanged);
			// 
			// label1
			// 
			this.label1.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.label1.Location = new System.Drawing.Point(104, 80);
			this.label1.Name = "label1";
			this.label1.Size = new System.Drawing.Size(88, 23);
			this.label1.TabIndex = 4;
			this.label1.Text = "Specular Intensity:";
			// 
			// trackIntensity
			// 
			this.trackIntensity.LargeChange = 20;
			this.trackIntensity.Location = new System.Drawing.Point(88, 40);
			this.trackIntensity.Maximum = 100;
			this.trackIntensity.Name = "trackIntensity";
			this.trackIntensity.Size = new System.Drawing.Size(120, 45);
			this.trackIntensity.SmallChange = 10;
			this.trackIntensity.TabIndex = 1;
			this.trackIntensity.TickFrequency = 10;
			this.trackIntensity.ValueChanged += new System.EventHandler(this.OnIntensityValueChanged);
			// 
			// lblIntensity
			// 
			this.lblIntensity.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.lblIntensity.Location = new System.Drawing.Point(100, 24);
			this.lblIntensity.Name = "lblIntensity";
			this.lblIntensity.Size = new System.Drawing.Size(88, 23);
			this.lblIntensity.TabIndex = 2;
			this.lblIntensity.Text = "Intensity:";
			// 
			// btnColour
			// 
			this.btnColour.BackColor = System.Drawing.Color.White;
			this.btnColour.Location = new System.Drawing.Point(16, 48);
			this.btnColour.Name = "btnColour";
			this.btnColour.Size = new System.Drawing.Size(56, 48);
			this.btnColour.TabIndex = 0;
			this.btnColour.Click += new System.EventHandler(this.OnColourButtonClick);
			// 
			// lblColour
			// 
			this.lblColour.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.lblColour.Location = new System.Drawing.Point(20, 32);
			this.lblColour.Name = "lblColour";
			this.lblColour.Size = new System.Drawing.Size(40, 23);
			this.lblColour.TabIndex = 0;
			this.lblColour.Text = "Colour:";
			// 
			// label2
			// 
			this.label2.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.label2.Location = new System.Drawing.Point(20, 136);
			this.label2.Name = "label2";
			this.label2.Size = new System.Drawing.Size(88, 23);
			this.label2.TabIndex = 12;
			this.label2.Text = "X Position:";
			// 
			// label3
			// 
			this.label3.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.label3.Location = new System.Drawing.Point(20, 168);
			this.label3.Name = "label3";
			this.label3.Size = new System.Drawing.Size(88, 23);
			this.label3.TabIndex = 13;
			this.label3.Text = "Y Position:";
			// 
			// label4
			// 
			this.label4.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.label4.Location = new System.Drawing.Point(20, 200);
			this.label4.Name = "label4";
			this.label4.Size = new System.Drawing.Size(88, 23);
			this.label4.TabIndex = 14;
			this.label4.Text = "Z Position:";
			// 
			// groupBox4
			// 
			this.groupBox4.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.groupBox4.Controls.Add(this.btnClose);
			this.groupBox4.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.groupBox4.Location = new System.Drawing.Point(-8, 264);
			this.groupBox4.Name = "groupBox4";
			this.groupBox4.Size = new System.Drawing.Size(336, 100);
			this.groupBox4.TabIndex = 99;
			this.groupBox4.TabStop = false;
			// 
			// btnClose
			// 
			this.btnClose.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.btnClose.DialogResult = System.Windows.Forms.DialogResult.OK;
			this.btnClose.FlatStyle = System.Windows.Forms.FlatStyle.System;
			this.btnClose.Location = new System.Drawing.Point(132, 20);
			this.btnClose.Name = "btnClose";
			this.btnClose.TabIndex = 0;
			this.btnClose.Text = "&Close";
			// 
			// colorDialog
			// 
			this.colorDialog.AnyColor = true;
			this.colorDialog.FullOpen = true;
			// 
			// LightsDialog
			// 
			this.AcceptButton = this.btnClose;
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(322, 317);
			this.Controls.Add(this.groupBox4);
			this.Controls.Add(this.groupBox3);
			this.Controls.Add(this.groupBox2);
			this.Controls.Add(this.groupBox1);
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedDialog;
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.MaximizeBox = false;
			this.MinimizeBox = false;
			this.Name = "LightsDialog";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterParent;
			this.Text = "Light Controls";
			this.groupBox1.ResumeLayout(false);
			this.groupBox2.ResumeLayout(false);
			this.groupBox3.ResumeLayout(false);
			((System.ComponentModel.ISupportInitialize)(this.trackSpecular)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.trackIntensity)).EndInit();
			this.groupBox4.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion
	

      // ------------------------------------------------------------------------------
      // Private data

      /// <summary>
      /// Registry keys
      /// </summary>
      public const string REG_LIGHT = "Light";
      public const string REG_LIGHT1 = "Light0";
      public const string REG_LIGHT2 = "Light1";
      public const string REG_LIGHT3 = "Light2";
      public const string REG_LIGHT4 = "Light3";

      /// <summary>
      /// The currently displayed light details
      /// </summary>
      private int m_currentLightIndex = 0;

      /// <summary>
      /// Default ambient light setting
      /// </summary>
      public static float[] LightAmbient  = new float[] { 0.0f, 0.0f, 0.0f, 1.0f };

      /// <summary>
      /// Controls
      /// </summary>
      private System.Windows.Forms.GroupBox groupBox1;
      private System.Windows.Forms.RadioButton radioLight1;
      private System.Windows.Forms.RadioButton radioLight2;
      private System.Windows.Forms.RadioButton radioLight3;
      private System.Windows.Forms.RadioButton radioLight4;
      private System.Windows.Forms.GroupBox groupBox2;
      private System.Windows.Forms.CheckBox checkLight1;
      private System.Windows.Forms.CheckBox checkLight2;
      private System.Windows.Forms.CheckBox checkLight4;
      private System.Windows.Forms.CheckBox checkLight3;
      private System.Windows.Forms.GroupBox groupBox3;
      private System.Windows.Forms.Label lblColour;
      private System.Windows.Forms.Label label1;
      private System.Windows.Forms.Label label2;
      private System.Windows.Forms.Label label3;
      private System.Windows.Forms.Label label4;
      private System.Windows.Forms.GroupBox groupBox4;
      private System.Windows.Forms.Button btnClose;
      private System.Windows.Forms.Button btnColour;
      private System.Windows.Forms.TrackBar trackSpecular;
      private System.Windows.Forms.HScrollBar hScrollBarXPos;
      private System.Windows.Forms.HScrollBar hScrollBarYPos;
      private System.Windows.Forms.HScrollBar hScrollBarZPos;
      private System.Windows.Forms.TrackBar trackIntensity;
      private System.Windows.Forms.Label lblIntensity;
		private System.Windows.Forms.ColorDialog colorDialog;
      private System.ComponentModel.Container components = null;
   }


   /// <summary>
   /// Describes a single OpenGL light instance
   /// </summary>
   public sealed class LightInfo
   {
      /// <summary>
      /// Save the light info to the Registry
      /// </summary>
      /// <param name="key"></param>
      public void SerialiseToRegistry(string key)
      {
         RegistryService.Instance.PutBoolean(key, REG_ACTIVE, Active);
         RegistryService.Instance.PutIntegerArray(key, REG_COLOUR, Colour.ToRGB());
         RegistryService.Instance.PutInteger(key, REG_INTENSITY, (int)(Intensity*FLOAT_TO_INT));
         RegistryService.Instance.PutInteger(key, REG_SPECULAR, (int)(Specular*FLOAT_TO_INT));
         RegistryService.Instance.PutIntegerArray(key, REG_POSITION, PointToIntArray(Position));
      }

      /// <summary>
      /// Restore the light info to the Registry
      /// </summary>
      /// <param name="key"></param>
      /// <returns>True if info found in registry, false otherwise</returns>
      public bool DeserialiseFromRegistry(string key)
      {
         if (RegistryService.Instance.KeyExists(key) == true)
         {
            Active    = RegistryService.Instance.GetBoolean(key, REG_ACTIVE, true);
            Colour    = new ColourTriple(RegistryService.Instance.GetIntegerArray(key, REG_COLOUR));
            Intensity = ((float)RegistryService.Instance.GetInteger(key, REG_INTENSITY, 1000)) / FLOAT_TO_INT;
            Specular  = ((float)RegistryService.Instance.GetInteger(key, REG_SPECULAR, 500)) / FLOAT_TO_INT;
            Position  = IntArrayToPoint(RegistryService.Instance.GetIntegerArray(key, REG_POSITION));
            
            return true;
         }
         else
         {
            return false;
         }
      }

      /// <summary>
      /// Convert a Point to an int[]
      /// </summary>
      /// <param name="point"></param>
      /// <returns></returns>
      private int[] PointToIntArray(K3D.Geometry.Point point)
      {
         return new int[3] { (int)(point.X*FLOAT_TO_INT), (int)(point.Y*FLOAT_TO_INT), (int)(point.Z*FLOAT_TO_INT) };
      }

      /// <summary>
      /// Convert an int[] to a Point
      /// </summary>
      /// <param name="vals"></param>
      /// <returns></returns>
      private K3D.Geometry.Point IntArrayToPoint(int[] vals)
      {
         if (vals != null && vals.Length == 3)
         {
            return new K3D.Geometry.Point( ((float)vals[0])/FLOAT_TO_INT, ((float)vals[1])/FLOAT_TO_INT, ((float)vals[2])/FLOAT_TO_INT );
         }
         else
         {
            return new K3D.Geometry.Point(0.0f, 0.0f, 0.0f);
         }
      }


      /// <summary>
      /// Public properties
      /// </summary>
      public bool Active;
      public ColourTriple Colour = new ColourTriple(1.0f, 1.0f, 1.0f);
      public float Intensity;
      public float Specular;
      public K3D.Geometry.Point Position = new K3D.Geometry.Point(0.0f, 0.0f, 0.0f);


      /// <summary>
      /// Private data
      /// </summary>
      private const float FLOAT_TO_INT    = 1000.0f;

      private const string REG_ACTIVE     = "active";
      private const string REG_COLOUR     = "colour";
      private const string REG_INTENSITY  = "intensity";
      private const string REG_SPECULAR   = "specular";
      private const string REG_POSITION   = "position";
   }
}
