namespace K3D.Services
{
	using System;
	using Microsoft.Win32;


	/// <summary>
	/// Registry Access Service.
	/// </summary>
	public sealed class RegistryService
	{
		// --------------------------------------------------------------------------------
		// Construction

		/// <summary>
		/// Private constructor
		/// </summary>
		private RegistryService()
		{
		}
		
		/// <summary>
		/// Static instance
		/// </summary>
		public static RegistryService Instance
		{
			get
			{
				return s_service;
			}
		}


		// --------------------------------------------------------------------------------
		// Public API

		/// <summary>
		/// Register the K3D-GL file types
		/// </summary>
		/// <param name="appPath">Full exec and app path string</param>
		public void RegisterFileTypes(string appPath)
		{
			RegistryKey filetypeRoot = Registry.ClassesRoot.CreateSubKey(".lwo");
			filetypeRoot.SetValue(null, "K3D.Document");
			filetypeRoot.Close();
         filetypeRoot = Registry.ClassesRoot.CreateSubKey(".LWO");
         filetypeRoot.SetValue(null, "K3D.Document");
         filetypeRoot.Close();
			
			RegistryKey k3dRoot = Registry.ClassesRoot.CreateSubKey("K3D.Document");
			k3dRoot.SetValue(null, "K3D Object File");
			RegistryKey iconKey = k3dRoot.CreateSubKey("DefaultIcon");
         string path = appPath.Substring(0, appPath.LastIndexOf('\\') + 1);
			iconKey.SetValue(null, path + "K3DGL.ico");
			RegistryKey openKey = k3dRoot.CreateSubKey("shell").CreateSubKey("open");
         openKey.SetValue(null, "Open with K3D...");
         RegistryKey cmdKey = openKey.CreateSubKey("command");
			cmdKey.SetValue(null, appPath + " \"%1\"");

			k3dRoot.Close();
		}

      /// <summary>
      /// Return whether a registry key exists
      /// </summary>
      /// <param name="key"></param>
      /// <returns></returns>
      public bool KeyExists(string key)
      {
         return (GetReadableKey(key) != null);
      }

		/// <summary>
		/// Get a string value from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public string GetString(string key, string name, string defaultVal)
		{
			string result = defaultVal;
         
			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				Object objValue = readableKey.GetValue(name, defaultVal);
				if (objValue != null)
				{
					result = objValue.ToString();
				}
			}
         
			return result;
		}

		/// <summary>
		/// Get an integer value from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public int GetInteger(string key, string name, int defaultVal)
		{
			int result = defaultVal;
         
			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				try
				{
					return Convert.ToInt32(readableKey.GetValue(name, defaultVal));
				}
				catch (FormatException)
				{
					// default will be returned
				}
			}

			return result;
		}

		/// <summary>
		/// Get a boolean value from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public bool GetBoolean(string key, string name, bool defaultVal)
		{
			bool result = defaultVal;

			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				try
				{
					return Convert.ToBoolean(readableKey.GetValue(name, defaultVal));
				}
				catch (FormatException)
				{
					// default will be returned
				}
			}
         
			return result;
		}

		/// <summary>
		/// Get an array of string values from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public string[] GetStringArray(string key, string name)
		{
			string[] result = null;

			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				result = readableKey.GetValue(name) as string[];
			}
         
			return result;
		}

		/// <summary>
		/// Get an array of int values from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public int[] GetIntegerArray(string key, string name)
		{
			int[] result = null;

			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				string[] strVals = readableKey.GetValue(name) as string[];
				if (strVals != null)
				{
					result = new int[strVals.Length];
					for (int i=0; i<strVals.Length; i++)
					{
						try
						{
							result[i] = Convert.ToInt32(strVals[i]);
						}
						catch (FormatException)
						{
							// default will be returned
						}
					}
				}
			}
         
			return result;
		}

		/// <summary>
		/// Get an array of byte values from a registry key
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="defaultVal">Default value to return if not set</param>
		/// <returns>Value</returns>
		public byte[] GetBinary(string key, string name)
		{
			byte[] result = null;

			RegistryKey readableKey = GetReadableKey(key);
			if (readableKey != null)
			{
				result = readableKey.GetValue(name) as byte[];
			}
         
			return result;
		}

		/// <summary>
		/// Put a string value into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">String value to put</param>
		public void PutString(string key, string name, string val)
		{
			SetValue(key, name, val);
		}

		/// <summary>
		/// Put a int value into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">Int value to put</param>
		public void PutInteger(string key, string name, int val)
		{
			SetValue(key, name, val);
		}

		/// <summary>
		/// Put a boolean value into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">Boolean value to put</param>
		public void PutBoolean(string key, string name, bool val)
		{
			SetValue(key, name, val);
		}

		/// <summary>
		/// Put a string array into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">String[] value to put</param>
		public void PutStringArray(string key, string name, string[] val)
		{
			SetValue(key, name, val);
		}

		/// <summary>
		/// Put an int array into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">Int[] value to put</param>
		public void PutIntegerArray(string key, string name, int[] val)
		{
			string[] strVals = new string[val.Length];
			for (int i=0; i<strVals.Length; i++)
			{
				try
				{
					strVals[i] = val[i].ToString();
				}
				catch (FormatException)
				{
					// default will be returned
				}
			}
			SetValue(key, name, strVals);
		}

		/// <summary>
		/// Put a byte array into a named registry ket
		/// </summary>
		/// <param name="key">Key - can be multiple separated by slash '/' characters</param>
		/// <param name="name">Name of the value within the key</param>
		/// <param name="val">Byte[] value to put</param>
		public void PutBinary(string key, string name, byte[] val)
		{
			SetValue(key, name, val);
		}


		// --------------------------------------------------------------------------------
		// Private methods

		/// <summary>
		/// Return a RegistryKey that is writable
		/// </summary>
		/// <param name="key">Key name</param>
		/// <returns>RegistryKey</returns>
		private RegistryKey GetWritableKey(string key)
		{
			RegistryKey root = Registry.CurrentUser.CreateSubKey(ROOT_KEY);
			return root.CreateSubKey(key);
		}

		/// <summary>
		/// Return a RegistryKey that is readonly
		/// </summary>
		/// <param name="key">Key name</param>
		/// <returns>RegistryKey</returns>
		private RegistryKey GetReadableKey(string key)
		{
			RegistryKey root = Registry.CurrentUser.CreateSubKey(ROOT_KEY);
			return root.OpenSubKey(key);
		}

		/// <summary>
		/// Set a value into the Register
		/// </summary>
		/// <param name="key"></param>
		/// <param name="name"></param>
		/// <param name="val"></param>
		private void SetValue(string key, string name, object val)
		{
			RegistryKey writableKey = GetWritableKey(key);
			writableKey.SetValue(name, val);
		}


		// --------------------------------------------------------------------------------
		// Private data
      
		private const string ROOT_KEY = @"Software\Twisted Wave\K3DGL2";

		private static RegistryService s_service = new RegistryService();
	}
}
