namespace K3D.ThreeD.Lightwave
{
	using System;
	using System.Diagnostics;
	using System.Collections;

   using K3D.IO.IFF;
   using K3D.Geometry;
	using K3D.Services;


	/// <summary>
	/// Visitor for LWO2 (Lightwave versions 6.0 to 7.5) format file.
	/// </summary>
	public sealed class LWO2Visitor : IChunkVisitor
	{
		// ------------------------------------------------------------------------------
		// Construction
      
		/// <summary>
		/// Constructor
		/// </summary>
		public LWO2Visitor()
		{
			m_lwObject = new LightwaveObject();
		}
      
      
		// ------------------------------------------------------------------------------
		// Interface implementation
      
		/// <summary>
		/// Deal with a single IFFChunk object.
		/// It is the responsibility of the vistor object to either Skip() the remainder
		/// of the chunk or GetData() as appropriate for this chunk type.
		/// </summary>
		/// <param name="chunk">An IFFChunk object</param>
		/// <param name="strId">The 4 character ID of the chunk</param>
		/// <param name="length">The length of the chunk in bytes</param>
		/// <returns>A ChunkReturn value as appropriate.</returns>
		public ChunkReturn VisitChunk(IFFChunk chunk, string strId, int length)
		{
			//if (K3DGLForm.DEBUG) Trace.WriteLine(strId + " length: " + length);
         
			// default return value is too skip this chunk if not a known ID
			ChunkReturn returnValue = ChunkReturn.Skip;
			switch (strId)
			{
				//case "TAGS":
				//	ReadTags(chunk.GetData());
				//	returnValue = ChunkReturn.ReadNextChunk;
				//	break;

				case "LAYR":
					StartLayer(chunk.GetData());
					returnValue = ChunkReturn.ReadNextChunk;
					break;

				case "PNTS":
					ReadPoints(chunk.GetData());
					returnValue = ChunkReturn.ReadNextChunk;
					break;

				case "POLS":
					ReadPolygons(chunk.GetData());
					returnValue = ChunkReturn.ReadNextChunk;
					break;

				case "PTAG":
					ReadPolygonTags(chunk.GetData());
					returnValue = ChunkReturn.ReadNextChunk;
					break;

				case "SURF":
               ReadSurface(chunk.GetData());
               returnValue = ChunkReturn.ReadNextChunk;
					break;
			}
         
			return returnValue;
		}

		/// <summary>
		/// Informs the visitor that all chunks have been processed
		/// </summary>
		public void EndVisit()
		{
			if (CurrentLayer != null)
			{
				SubPatchPolygons();
			}

			// validate that each of the layers has polygons to display
			ArrayList temp = (ArrayList)m_layers.Clone();
			foreach (LightwaveLayer layer in temp)
			{
				if (layer.Polygons == null)
				{
					m_layers.Remove(layer);
				}
			}
			m_lwObject.Layers = (LightwaveLayer[])m_layers.ToArray(typeof(LightwaveLayer));

			// set the surfaces
			if (m_surfaces.Count == 0)
			{
				// add default surface if none loaded
				m_surfaces.Add(new LightwaveSurface());
			}
			m_lwObject.Surfaces = (LightwaveSurface[])m_surfaces.ToArray(typeof(LightwaveSurface));

			// set the object info
			foreach (LightwaveLayer layer in m_lwObject.Layers)
			{
				m_lwObject.ObjectInfo.NumPolygons += layer.Polygons.Length;
				m_lwObject.ObjectInfo.NumPoints += layer.Points.Length;
			}
			m_lwObject.ObjectInfo.NumSurfaces = m_lwObject.Surfaces.Length;
			m_lwObject.ObjectInfo.NumLayers = m_lwObject.Layers.Length;
		}

      /// <summary>
      /// Return the 3D object created by the visitor
      /// </summary>
      /// <returns>ThreeDObject or null if not created yet</returns>
      public ThreeDObject Get3DObject()
      {
         return m_lwObject;
      }
      
      
		// ------------------------------------------------------------------------------
		// Private methods
      
		/// <summary>
		/// Start a new object layer
		/// </summary>
		/// <param name="data"></param>
		private void StartLayer(byte[] data)
		{
         // TODO: handle 'hidden' layers?
			if (CurrentLayer != null)
			{
				SubPatchPolygons();
			}
			
			if (K3DGLForm.DEBUG) Trace.WriteLine("Starting object layer: " + (m_layers.Count+1));
			m_layers.Add(new LightwaveLayer());
		}

		/// <summary>
		/// The currently active layer
		/// </summary>
		private LightwaveLayer CurrentLayer
		{
			get
			{
				if (m_layers.Count != 0)
				{
					return (LightwaveLayer)m_layers[m_layers.Count - 1];
				}
				else
				{
					return null;
				}
			}
		}

		/// <summary>
		/// Read the points values from the specified data block
		/// </summary>
		/// <param name="data">raw byte data containing point values</param>
		private void ReadPoints(byte[] data)
		{
			if (CurrentLayer == null)
			{
				throw new IFFReaderException("Points chunk found before layer!");
			}

			if (K3DGLForm.DEBUG) Trace.WriteLine("Read PNTS chunk containing " + (data.Length/12) + " points");
			
			int iPointTotal = data.Length / 12;
			Point[] points = new Point[iPointTotal];
			
			Point point;
			int iOffset = 0;
			for (int iCount=0; iCount<iPointTotal; iCount++)
			{
				point = new Point(IFFChunk.ReadIEEEfloat32(data, ref iOffset),
										IFFChunk.ReadIEEEfloat32(data, ref iOffset),
										IFFChunk.ReadIEEEfloat32(data, ref iOffset));
            
				//if (K3DGLForm.DEBUG) Trace.WriteLine(point.ToString());
				
				points[iCount] = point;
			}

			// set points on current layer
			CurrentLayer.Points = points;
		}
      
		/// <summary>
		/// Read the polygon values from the specified data block. Uses the most recent
		/// points values to form the polygon objects.
		/// </summary>
		/// <param name="data">raw byte data containing point values</param>
		private void ReadPolygons(byte[] data)
		{
			string strTypeId = IFFChunk.ReadID(data);
			if (strTypeId == "FACE" || strTypeId == "PTCH")
			{
				if (CurrentLayer == null || CurrentLayer.Points == null)
				{
					throw new IFFReaderException("Points chunk not found before Polygons chunk!");
				}

				// guess reasonable size for polygon list
				if (K3DGLForm.DEBUG) Trace.WriteLine("Guessing poly size array of: " + (CurrentLayer.Points.Length + (CurrentLayer.Points.Length>>3)));
				ArrayList polygons = new ArrayList(CurrentLayer.Points.Length + (CurrentLayer.Points.Length>>3));
				
				int iOffset = 4;  // skip id
				while (iOffset < data.Length)
				{
					// calc number of vertices (mask top 6 flag bits)
					int iNumVerts = IFFChunk.ReadInt16(data, ref iOffset) & 0x00003fff;
               
					// build vertex list required for this polygon
					int[] vectors = new int[iNumVerts];
					for (int iPointCount=0; iPointCount<iNumVerts; iPointCount++)
					{
						int iVertex;
						// check variable length index type
						if (data[iOffset] != (byte)0xff)
						{
							iVertex = IFFChunk.ReadInt16(data, ref iOffset);
						}
						else
						{
							// mask top byte
							data[iOffset] = 0;
							iVertex = IFFChunk.ReadInt32(data, ref iOffset);
						}
                  
						// reference the specified Point for our polygon
						vectors[iPointCount] = iVertex;
					}
               
					// create polygon from points
					Polygon polygon = new Polygon(vectors);
					polygons.Add(polygon);

					//if (K3DGLForm.DEBUG) Trace.WriteLine(polygon.ToString());
				}
				
				// set polygons on current layer
				// NOTE: even with multiple POLS chunks e.g. both a FACE and PTCH
				// in the same layer, there is never more than 1 of each chunk type
				if (strTypeId == "FACE")
				{
					CurrentLayer.RawPolygons = polygons;
				}
				else
				{
					CurrentLayer.SubPatchPolygons = polygons;
				}
				
				if (K3DGLForm.DEBUG) Trace.WriteLine("Read POLS chunk (" + strTypeId + ") containing " + polygons.Count + " polys");
			}
		}

		/// <summary>
		/// Use the Catmull-Clark subdivision algorithm to perform "fake"
		/// subpatching on our polygons.
		/// NOTE: this must be performed after ALL data has been read for the current
		///		 layer as a new layer may be created to hold non-subpatch polys
		/// </summary>
		private void SubPatchPolygons()
		{
			if (CurrentLayer.SubPatchPolygons == null) return;

			LightwaveLayer otherLayer = null;
			if (CurrentLayer.RawPolygons != null)
			{
				// prepare another layer for the non-patch data so we can start afresh
				if (CurrentLayer.RawPolygons != null)
				{
					// other poly data found, so copy out old data to new layer
					otherLayer = new LightwaveLayer();
					otherLayer.RawPolygons = CurrentLayer.RawPolygons;
					CurrentLayer.RawPolygons = null;
					otherLayer.Points = (Point[])CurrentLayer.Points.Clone();
				}
			}
			
			// TODO: read and apply weight-maps for the points!
			if (PropertiesService.Instance.Subpatch == true)
			{
				ArrayList polygons = CurrentLayer.SubPatchPolygons;
			
				// repeat as per sub-patch level from user Properties
				int patchLevel = PropertiesService.Instance.PatchLevel;
				for (int patchCount=0; patchCount<patchLevel; patchCount++)
				{
					Point[] points = CurrentLayer.Points;
					
					int totalPolys = polygons.Count;
					int totalPoints = points.Length;
					
					PointInfo[] pointInfos = new PointInfo[totalPoints];
					
					// TODO: check how big this structure grows!
					EdgeMap edgeMap = new EdgeMap(totalPolys << 1);
					
					// TODO: should only need to do this once, it should be possible
					//		   to re-make it as we generate new polys during the algorithm
					//if (patchCount == 0)
					//{
					// populate edgemap for the first time
					LightwaveObject.PopulatePolygonEdgeMap(polygons, edgeMap);
					//}
					
					ArrayList newPolygons = new ArrayList(totalPolys << 2);
					ArrayList newPoints = new ArrayList(totalPoints << 2);
					if (K3DGLForm.DEBUG) Trace.WriteLine("Guessing size of New Poly and Points lists: " + newPolygons.Capacity + " & " + newPoints.Capacity);
					
					// first pass we generate new face and new edges points for each poly
					// we also store references to those against each vertex in the poly
					for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
					{
						// generate face point for poly
						newPoints.Add( LightwaveObject.PolygonMidPoint((Polygon)polygons[polyIndex], points) );
					}
					
					// second pass we generate averaged edge points for each edge in a poly
					// the edge points are the edge vertex and adjacent face points averages
					// the new edge points shared by polys must continue to be shared - otherwise
					//	the polys "break apart" on the 2nd+ iteration.
					for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
					{
						Polygon poly = (Polygon)polygons[polyIndex];
						int vertexCount = poly.VertexCount;
						
						// generate edge point for each vertex pair
						for (int vertexIndex=0; vertexIndex<vertexCount; vertexIndex++)
						{
							int edgeVertex1 = poly.Vertices[vertexIndex];
							Point vertex = points[ edgeVertex1 ];
							
							int edgeVertex2;
							if (vertexIndex < vertexCount - 1)
							{
								edgeVertex2 = poly.Vertices[vertexIndex + 1];
							}
							else
							{
								edgeVertex2 = poly.Vertices[0];
							}
							Point nextPoint = points[ edgeVertex2 ];
							
							// store an index to this edgepoint (and facepoint) for the Point
							if (pointInfos[edgeVertex1] == null)
							{
								pointInfos[edgeVertex1] = new PointInfo();
							}
							if (pointInfos[edgeVertex2] == null)
							{
								pointInfos[edgeVertex2] = new PointInfo();
							}

							// get edge info for the edge we are working on
							EdgeInfo edgeInfo = edgeMap[edgeVertex1, edgeVertex2];
							// we only add the edge/mid-point for an edge once
							if (edgeInfo.EdgePointIndex == -1)
							{
								Point plainEdgePoint = LightwaveObject.EdgeMidPoint(vertex, nextPoint);
								pointInfos[edgeVertex1].AddEdgePoint(plainEdgePoint);
								pointInfos[edgeVertex2].AddEdgePoint(plainEdgePoint);
							}
							
							// store the face point for this vertex once only
							pointInfos[edgeVertex1].AddFacePointRef(polyIndex);
							
							// we only add the edge/mid-point for an edge once
							if (edgeInfo.EdgePointIndex == -1)
							{
								// get adjacent face points and combine them into the average
								Point facePoint = (Point)newPoints[ polyIndex ];
								Point edgePoint;
								int poly2Index = edgeInfo.PolygonIndex1;
								if (polyIndex == poly2Index)
								{
									poly2Index = edgeInfo.PolygonIndex2;
								}
								if (poly2Index != -1)
								{
									Point facePoint2 = (Point)newPoints[ poly2Index ];
									edgePoint = new Point(
										(vertex.X + nextPoint.X + facePoint.X + facePoint2.X) * 0.25f,
										(vertex.Y + nextPoint.Y + facePoint.Y + facePoint2.Y) * 0.25f,
										(vertex.Z + nextPoint.Z + facePoint.Z + facePoint2.Z) * 0.25f);
								}
								else
								{
									// no second poly found - we must be at a boundry
									edgePoint = new Point(
										(vertex.X + nextPoint.X + facePoint.X) * 0.33333f,
										(vertex.Y + nextPoint.Y + facePoint.Y) * 0.33333f,
										(vertex.Z + nextPoint.Z + facePoint.Z) * 0.33333f);
								}
								
								// store the new edgepoint
								edgeInfo.EdgePointIndex = newPoints.Count;
								newPoints.Add(edgePoint);
							}
						}
					}
					
					// third pass we modify the original vertices to new positions
					// based on the Catmull Clark algorithm and store the new points
					for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
					{
						Polygon poly = (Polygon)polygons[polyIndex];
						int vertexCount = poly.VertexCount;
						
						for (int vertexIndex=0; vertexIndex<vertexCount; vertexIndex++)
						{
							// we recalc the vertex at the end of the current edge
							int edgeVertex2;
							if (vertexIndex < vertexCount - 1)
							{
								edgeVertex2 = poly.Vertices[vertexIndex + 1];
							}
							else
							{
								edgeVertex2 = poly.Vertices[0];
							}
							
							// deal with point at end of the edge
							Point point = points[ edgeVertex2 ];
							PointInfo pointInfo = pointInfos[ edgeVertex2 ];
							
							// only calc vertex points for an edge once
							//	then use those references when creating the new polys
							if (pointInfo.NewVertexPointIndex == -1)
							{
								// modify point pos according to Catmull Clark algorithm
								//     Q  +  2R  +  S(n-3)
								//    --------------------
								//		         n       
								// Q is the average of the new face points surrounding the old vertex, 
								// R is the average of the midpoints of the edges that share the old vertex, 
								// S is the old vertex point, and 
								// n is the number of edges that share the old vertex. 
								
								float n = (float)(pointInfo.EdgePointCount);
								float sn3 = (n - 3.0f);
								float mult = 1.0f / n;
								
								// (the average of the face points for those adjoining polygons 
								Point pointQ = pointInfo.FacePointAverage(newPoints);
								
								// the average of the midpoints of the edges that touch the old vertex
								Point point2R = pointInfo.EdgePointAverage2R(newPoints);
								
								// (n - 3) times the old vertex
								Point pointSN3 = new Point(point.X * sn3,
																	point.Y * sn3,
																	point.Z * sn3);
								
								Point average = new Point(
									(pointQ.X + point2R.X + pointSN3.X) * mult,
									(pointQ.Y + point2R.Y + pointSN3.Y) * mult,
									(pointQ.Z + point2R.Z + pointSN3.Z) * mult);
								
								// save the index to the recalculated vertex point
								pointInfo.NewVertexPointIndex = newPoints.Count;
								
								newPoints.Add(average);
							}
						}
						
						// create new polys now by connecting up points
						for (int vertexIndex=0; vertexIndex<vertexCount; vertexIndex++)
						{
							// work out the vertex indexes for the two edges we need
							int edgeVertex1 = poly.Vertices[vertexIndex];
							int edgeVertex2;
							if (vertexIndex < vertexCount - 1)
							{
								edgeVertex2 = poly.Vertices[vertexIndex + 1];
							}
							else
							{
								edgeVertex2 = poly.Vertices[0];
							}
							int edgeVertex3;
							if (vertexIndex < vertexCount - 2)
							{
								edgeVertex3 = poly.Vertices[vertexIndex + 2];
							}
							else if (vertexIndex < vertexCount - 1)
							{
								edgeVertex3 = poly.Vertices[0];
							}
							else
							{
								edgeVertex3 = poly.Vertices[1];
							}
							
							// get the two edges we need
							EdgeInfo edgeInfo1 = edgeMap[edgeVertex1, edgeVertex2];
							EdgeInfo edgeInfo2 = edgeMap[edgeVertex2, edgeVertex3];
							
							// we connect to the vertex at the end point of the first edge
							PointInfo vertexEndPointInfo = pointInfos[edgeVertex2];
							
							int facePointIndex = polyIndex;
							
							// Catmull-Clark scheme always produces quads
							int[] polyVerts = new int[4];
							polyVerts[0] = facePointIndex;						// face point
							polyVerts[1] = edgeInfo1.EdgePointIndex;			// first edge point
							polyVerts[2] = vertexEndPointInfo.NewVertexPointIndex;
							polyVerts[3] = edgeInfo2.EdgePointIndex;			// second edge point
							
							// TODO: add to new edge list here!
							Polygon newPoly = new Polygon(polyVerts);
							
							// force smoothing on the surface of this polygon
							newPoly.SurfaceIndex = poly.SurfaceIndex;
							newPoly.ForceSmoothing = true;
							newPolygons.Add(newPoly);
						}
					}
					
					if (K3DGLForm.DEBUG) Trace.WriteLine("Real size of New Poly and Points lists: " + newPolygons.Count + " & " + newPoints.Count);

					// overwrite current points with new points
					CurrentLayer.Points = (Point[])newPoints.ToArray(typeof(Point));
					polygons = newPolygons;
					
					if (patchCount + 1 == patchLevel)
					{
						CurrentLayer.SubPatchPolygons = newPolygons;
					}
				}
			 
				if (K3DGLForm.DEBUG) Trace.WriteLine("Converted to sub-patches as " + CurrentLayer.SubPatchPolygons.Count + " polygons and " + CurrentLayer.Points.Length + " points.");
			}

			// insert our other layer if required
			if (otherLayer != null)
			{
				m_layers.Add(otherLayer);
			}
		}

		/// <summary>
		/// Read the string tag values for this object
		/// </summary>
		/// <param name="data"></param>
		private void ReadTags(byte[] data)
		{
			ArrayList tags = new ArrayList();
			int offset = 0;
			while (offset < data.Length - 1)
			{
				string tag = IFFChunk.ReadNullTerminatedString(data, ref offset);
				tags.Add( tag );
			}
			m_lwObject.Tags = (string[])tags.ToArray(typeof(string));

			if (K3DGLForm.DEBUG) Trace.WriteLine("Read TAGS chunk containing " + tags.Count + " tag strings");
		}

      /// <summary>
      /// Read a surface block
      /// </summary>
      /// <param name="data"></param>
      private void ReadSurface(byte[] data)
      {
         LightwaveSurface surface = new LightwaveSurface();

         int offset = 0;

         // read each subchunk, setting the appropriate surface attributes
         // as we fund them in recognised chunks
         string name = IFFChunk.ReadNullTerminatedString(data, ref offset);
         string parent = IFFChunk.ReadNullTerminatedString(data, ref offset);
         
         while (offset < data.Length - 1)
         {
            string subchunkId = IFFChunk.ReadID(data, ref offset);
            int subchunkLength = IFFChunk.ReadInt16(data, ref offset);
            if (subchunkLength % 2 == 1)
            {
               subchunkLength++;
            }
            int nextChunkOffset = offset + subchunkLength;
            
            switch (subchunkId)
            {
               case "COLR":
                  surface.Colour = IFFChunk.ReadColourTriple(data, ref offset);
                  break;
               case "DIFF":
                  surface.Diffusion = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "LUMI":
                  surface.Luminosity = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "SPEC":
                  surface.Specular = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "TRAN":
                  surface.Transparency = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "GLOS":
                  surface.Glossiness = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "SMAN":
						// in radians for LW6 format
                  surface.MaxSmoothing = IFFChunk.ReadIEEEfloat32(data, ref offset);
						if (surface.MaxSmoothing > 0.0f)
						{
							surface.Smoothing = true;
						}
                  break;
					case "SIDE":
						// 1 is front only, 3 is front & back
						surface.DoubleSided = (IFFChunk.ReadInt16(data, ref offset) == 3);
						break;
               case "BLOK":
                  // texture/procedural mapping block found
                  //ReadTextureBlock(data, ref offset, surface);
                  break;
               default:
                  // unknown chunk, ignore it
                  break;
            }
            
            // inc offset ready for next chunk
            offset = nextChunkOffset;
         }
         
         // add surface to list
         m_surfaces.Add(surface);
      }

      /// <summary>
      /// Read a texture block from the ata into the surface
      /// </summary>
      /// <param name="data">byte[] containing the block</param>
      /// <param name="offset">Offset to start of block</param>
      /// <param name="surface">Surface to add info to</param>
      private void ReadTextureBlock(byte[] data, ref int offset, LightwaveSurface surface)
      {
         string id = IFFChunk.ReadID(data, ref offset);
         if (id == "IMAP")
         {
            // found an imagemap block
            int length = IFFChunk.ReadInt16(data, ref offset);
            string ordinal = IFFChunk.ReadNullTerminatedString(data, ref offset);
            int pos = 0;
            while (pos < length)
            {
               // read sub-chunks that we are interested in
               string subId = IFFChunk.ReadID(data, ref offset);
					int subLength = IFFChunk.ReadInt16(data, ref offset);
					switch (subId)
					{
						default:
							break;
					}
					pos += subLength;
            }
         }
      }

		/// <summary>
		/// Read the polygon tag mappings for this layer
		/// </summary>
		/// <param name="data"></param>
		private void ReadPolygonTags(byte[] data)
		{
			if (CurrentLayer == null)
			{
				throw new IFFReaderException("Polygon tags chunk found before layer!");
			}

			if (CurrentLayer.LastAddedPolygons != null)
			{
				int offset = 0;
				string strTypeId = IFFChunk.ReadID(data, ref offset);
				int count = 0;
				if (strTypeId == "SURF")
				{
					try
					{
						while (offset < data.Length - 1)
						{
							// surface mappings found
							int polyIndex = IFFChunk.ReadVariableLengthIndex(data, ref offset);
							int tagIndex = IFFChunk.ReadInt16(data, ref offset);
							if (tagIndex > 32767) 
							{
								// detail polys present???
								tagIndex = 65536 - tagIndex;
								
								// TODO: do we need this for LW6 format?
								((Polygon)CurrentLayer.LastAddedPolygons[polyIndex]).Ignore = true;
							}
							
							//Debug.Assert(polyIndex < CurrentLayer.Polygons.Length, "PTAG index references a non-existing polygon!");
							// set the surface index for the appropriate polygon
							((Polygon)CurrentLayer.LastAddedPolygons[polyIndex]).SurfaceIndex = tagIndex;
							
							count++;
						}
					}
					catch (Exception e)
					{
						Trace.WriteLine("ERROR: ReadPolygonTags poly count: " + count + " at offset: " + offset);
						throw e;
					}
				}
				// TODO: do we need this chunk?
				/*else if (strTypeId == "SMGP")
				{
					while (offset < data.Length - 1)
					{
						// smoothing group mappings found

						// offset is passed as a ref so it is incremented the correct ammount
						int polyIndex = IFFChunk.ReadVariableLengthIndex(data, ref offset);
						int tagIndex = IFFChunk.ReadInt16(data, ref offset);

						Debug.Assert(polyIndex < CurrentLayer.Polygons.Length, "PTAG index references a non-existing polygon!");
						// set the smoothing group index for the appropriate polygon
						CurrentLayer.Polygons[polyIndex + CurrentLayer.PolygonOffset].SmoothingGroupIndex = tagIndex;

						count++;
					}
				}*/
				if (K3DGLForm.DEBUG) Trace.WriteLine("Read PTAG chunk containing " + count + " " + strTypeId + " mappings");
			}
			else
			{
				if (K3DGLForm.DEBUG) Trace.WriteLine("WARNING: PTAGS chunk found before POLS chunk!");
			}
		}
      
      
		// ------------------------------------------------------------------------------
		// Private data
      
		/// <summary>
		/// The LightwaveObject we have created by visiting the IFF file
		/// </summary>
		private LightwaveObject m_lwObject = null;
      
		/// <summary>
		/// Holds the Layers that have been built so far
		/// </summary>
		private ArrayList m_layers = new ArrayList(10);

      /// <summary>
      /// Holds the surface definitions found so far
      /// </summary>
      private ArrayList m_surfaces = new ArrayList(16);
	}
}
