namespace K3D.ThreeD.Lightwave
{
   using System;
   using System.Diagnostics;
   using System.Collections;

   using K3D.IO.IFF;
   using K3D.Geometry;


   /// <summary>
   /// Visitor for LWOB (Lightwave versions 3.0 to 5.6) format file.
   /// </summary>
   public sealed class LWOBVisitor : IChunkVisitor
   {
      // ------------------------------------------------------------------------------
      // Construction
      
      /// <summary>
      /// Constructor
      /// </summary>
      public LWOBVisitor()
      {
         m_lwObject = new LightwaveObject();
      }
      
      
      // ------------------------------------------------------------------------------
      // Interface implementation
      
      /// <summary>
      /// Deal with a single IFFChunk object.
      /// It is the responsibility of the vistor object to either Skip() the remainder
      /// of the chunk or GetData() as appropriate for this chunk type.
      /// </summary>
      /// <param name="chunk">An IFFChunk object</param>
      /// <param name="strId">The 4 character ID of the chunk</param>
      /// <param name="length">The length of the chunk in bytes</param>
      /// <returns>A ChunkReturn value as appropriate.</returns>
      public ChunkReturn VisitChunk(IFFChunk chunk, string strId, int length)
      {
         //if (K3DGLForm.DEBUG) Trace.WriteLine(strId + " length: " + length);
         
         // default return value is too skip this chunk if not a known ID
         ChunkReturn returnValue = ChunkReturn.Skip;
         switch (strId)
         {
            case "PNTS":
               ReadPoints(chunk.GetData());
               returnValue = ChunkReturn.ReadNextChunk;
               break;

            case "POLS":
               ReadPolygons(chunk.GetData());
               returnValue = ChunkReturn.ReadNextChunk;
               break;

            case "SURF":
               ReadSurface(chunk.GetData());
               returnValue = ChunkReturn.ReadNextChunk;
               break;

            //case "SRFS":
            //   ReadTags(chunk.GetData());
            //   returnValue = ChunkReturn.ReadNextChunk;
            //   break;
         }
         
         return returnValue;
      }

      /// <summary>
      /// Informs the visitor that all chunks have been processed
      /// </summary>
      public void EndVisit()
      {
			// validate that each of the layers has polygons to display
			if (m_layers[0].Polygons != null)
			{
				m_lwObject.Layers = m_layers;
			}
			else
			{
				m_lwObject.Layers = new LightwaveLayer[0];
			}

			// set surfaces
			if (m_surfaces.Count == 0)
			{
				// add default surface if none loaded
				m_surfaces.Add(new LightwaveSurface());
			}
         m_lwObject.Surfaces = (LightwaveSurface[])m_surfaces.ToArray(typeof(LightwaveSurface));
         
			// set the object info
         if (m_lwObject.Layers != null && m_lwObject.Layers.Length != 0)
         {
            if (m_lwObject.Layers[0].Polygons != null)
            {
               m_lwObject.ObjectInfo.NumPolygons = m_lwObject.Layers[0].Polygons.Length;
            }
            m_lwObject.ObjectInfo.NumPoints = m_lwObject.Layers[0].Points.Length;
         }
			m_lwObject.ObjectInfo.NumSurfaces = m_lwObject.Surfaces.Length;
			m_lwObject.ObjectInfo.NumLayers = m_lwObject.Layers.Length;
      }

      /// <summary>
      /// Return the 3D object created by the visitor
      /// </summary>
      /// <returns>ThreeDObject or null if not created yet</returns>
      public ThreeDObject Get3DObject()
      {
         return m_lwObject;
      }
      
      
      // ------------------------------------------------------------------------------
      // Private methods

      /// <summary>
      /// The currently active layer
      /// </summary>
      private LightwaveLayer CurrentLayer
      {
         get
         {
            if (m_layers[0] == null)
            {
               m_layers[0] = new LightwaveLayer();
            }
            return m_layers[0];
         }
      }

      /// <summary>
      /// Read the points values from the specified data block
      /// </summary>
      /// <param name="data">raw byte data containing point values</param>
      private void ReadPoints(byte[] data)
      {
         if (K3DGLForm.DEBUG) Trace.WriteLine("Read PNTS chunk containing " + (data.Length/12) + " points");
			
         int iPointTotal = data.Length / 12;
         Point[] points = new Point[iPointTotal];
			
         Point point;
         int iOffset = 0;
         for (int iCount=0; iCount<iPointTotal; iCount++)
         {
            point = new Point(IFFChunk.ReadIEEEfloat32(data, ref iOffset),
               IFFChunk.ReadIEEEfloat32(data, ref iOffset),
               IFFChunk.ReadIEEEfloat32(data, ref iOffset));
            
            //if (K3DGLForm.DEBUG) Trace.WriteLine(point.ToString());
				
            points[iCount] = point;
         }

         // set points on current layer
         CurrentLayer.Points = points;
      }
      
      /// <summary>
      /// Read the polygon values from the specified data block. Uses the most recent
      /// points values to form the polygon objects.
      /// </summary>
      /// <param name="data">raw byte data containing point values</param>
      private void ReadPolygons(byte[] data)
      {
         if (CurrentLayer.Points == null)
         {
            throw new IFFReaderException("Points chunk not found before Polygons chunk!");
         }
         
			// guess reasonable size for polygon list
			if (K3DGLForm.DEBUG) Trace.WriteLine("Guessing poly size array of: " + CurrentLayer.Points.Length);
			ArrayList polygons = new ArrayList(CurrentLayer.Points.Length);
			
         int iOffset = 0;
         while (iOffset < data.Length)
         {
            // calc number of vertices for this polygon (mask top 6 flag bits)
            int iNumVerts = IFFChunk.ReadInt16(data, ref iOffset) & 0x00003fff;// 0x0000003f;
            
            // build vertex list required for this polygon
            int[] vectors = new int[iNumVerts];
            for (int iPointCount=0; iPointCount<iNumVerts; iPointCount++)
            {
               int iVertex = IFFChunk.ReadInt16(data, ref iOffset);
               
               // reference the specified Point for our polygon
               vectors[iPointCount] = iVertex;
            }
            
            // create polygon from points
            Polygon polygon = new Polygon(vectors);
            polygons.Add(polygon);

            // set surface index
            int surfaceIndex = IFFChunk.ReadInt16(data, ref iOffset);
				if (surfaceIndex > 32767) 
				{
					// detail polys present!
					// store correct surface info for this poly before we worry about detail polys
					polygon.SurfaceIndex = (65536 - surfaceIndex - 1);
					
					// skip detail polys
					// next byte is number of detail polys in list
					int detailPolyTotal = IFFChunk.ReadInt16(data, ref iOffset);
					for (int detailPolyCount=0; detailPolyCount<detailPolyTotal; detailPolyCount++)
					{
						// calc number of vertices for this polygon (mask top 6 flag bits)
						int iNumVertDetail = IFFChunk.ReadInt16(data, ref iOffset) & 0x00003fff;// 0x0000003f;
						
						// skip vertex list for detail poly (2 bytes per vertex)
						iOffset += (iNumVertDetail << 1);
						
						// skip surface index for detail poly
						iOffset += 2;
					}
				}
				else
				{
               polygon.SurfaceIndex = surfaceIndex - 1;
            }
				
            //if (K3DGLForm.DEBUG) Trace.WriteLine(polygon.ToString());
         }

         // set polygons on current layer
         CurrentLayer.RawPolygons = polygons;

         if (K3DGLForm.DEBUG) Trace.WriteLine("Read POLS chunk containing " + polygons.Count + " polys");
      }

      /// <summary>
      /// Read a surface block
      /// </summary>
      /// <param name="data"></param>
      private void ReadSurface(byte[] data)
      {
         LightwaveSurface surface = new LightwaveSurface();

         int offset = 0;

         // read each subchunk, setting the appropriate surface attributes
         // as we fund them in recognised chunks
         string name = IFFChunk.ReadNullTerminatedString(data, ref offset);
         
         while (offset < data.Length - 1)
         {
            string subchunkId = IFFChunk.ReadID(data, ref offset);
            int subchunkLength = IFFChunk.ReadInt16(data, ref offset);
            if (subchunkLength % 2 == 1)
            {
               subchunkLength++;
            }
            int nextChunkOffset = offset + subchunkLength;
            
            switch (subchunkId)
            {
               case "COLR":
                  surface.Colour = IFFChunk.ReadColourTripleByte(data, ref offset);
                  break;
               case "DIFF":
                  surface.Diffusion = ((float)IFFChunk.ReadInt16(data, ref offset)) * ColourTriple.RGB_BYTE_TO_FLOAT;
                  break;
               case "LUMI":
                  surface.Luminosity = ((float)IFFChunk.ReadInt16(data, ref offset)) * ColourTriple.RGB_BYTE_TO_FLOAT;
                  break;
               case "SPEC":
                  surface.Specular = ((float)IFFChunk.ReadInt16(data, ref offset)) * ColourTriple.RGB_BYTE_TO_FLOAT;
                  break;
               case "TRAN":
                  surface.Transparency = ((float)IFFChunk.ReadInt16(data, ref offset)) * ColourTriple.RGB_BYTE_TO_FLOAT;
                  break;
               case "GLOS":
                  // max of 1024
                  surface.Glossiness = ((float)IFFChunk.ReadInt16(data, ref offset)) / 256.0f;
                  break;
               case "SMAN":
                  // in radians for LW6 format
                  surface.MaxSmoothing = IFFChunk.ReadIEEEfloat32(data, ref offset);
                  break;
               case "FLAG":
                  // smoothing mode and double sided flags
                  int flags = IFFChunk.ReadInt16(data, ref offset);
                  surface.DoubleSided = ((flags & FLAGS_DOUBLESIDED) > 0);
						surface.Smoothing = ((flags & FLAGS_SMOOTHING) > 0);
                  break;
               default:
                  // unknown chunk, ignore it
                  break;
            }
            
            // inc offset ready for next chunk
            offset = nextChunkOffset;
         }

         // add surface to list
         m_surfaces.Add(surface);
      }

      /// <summary>
      /// Read the string tag values for this object
      /// </summary>
      /// <param name="data"></param>
      private void ReadTags(byte[] data)
      {
         ArrayList tags = new ArrayList();
         int offset = 0;
         while (offset < data.Length - 1)
         {
            string tag = IFFChunk.ReadNullTerminatedString(data, ref offset);
            tags.Add( tag );
         }
         m_lwObject.Tags = (string[])tags.ToArray(typeof(string));
         
         if (K3DGLForm.DEBUG) Trace.WriteLine("Read SRFS chunk containing " + tags.Count + " tag strings");
      }


      // ------------------------------------------------------------------------------
      // Private data
      
      private const int FLAGS_DOUBLESIDED = 1 << 8;	// if set then double sided surface
      private const int FLAGS_SMOOTHING = 1 << 2;		// if set then smooth polygon shading

      /// <summary>
      /// The LightwaveObject we have created by visiting the IFF file
      /// </summary>
      private LightwaveObject m_lwObject = null;
      
      /// <summary>
      /// Holds the Layers that have been built so far
      /// </summary>
      private LightwaveLayer[] m_layers = new LightwaveLayer[1];

      /// <summary>
      /// Holds the surface definitions found so far
      /// </summary>
      private ArrayList m_surfaces = new ArrayList(16);
   }
}
