namespace K3D.ThreeD
{
   using System;
	using System.Collections;
   using System.Diagnostics;

   using K3D.Geometry;


   /// <summary>
   /// ThreeDObject
   /// Base class for all 3D object types.
   /// </summary>
   public abstract class ThreeDObject
   {
      // ------------------------------------------------------------------------------
      // Public methods
      
      /// <summary>
      /// Centre and scale the object to the window size and origin
      /// </summary>
      public abstract void CentreAndScale();

      /// <summary>
      /// Calculate face (facet) normals for a 3D object
      /// Normals are inverted to account for reversed Z coords
      /// </summary>
      public abstract void CalculateFacetNormals();

      /// <summary>
      /// Calculate averaged vertex normals for a 3D object.
      /// This allows gouraud shading to be applied across polygons.
      /// For each finding all the polys that each point is referenced in,
      /// build a list of those references and average the normals
      /// </summary>
      public abstract void CalculateVertexNormals();

      /// <summary>
      /// Convert the object data to OpenGL display lists
      /// </summary>
      public abstract void BuildOpenGLDisplayLists();

      /// <summary>
      /// Return the number of OpenGL display lists this object requires
      /// </summary>
      public abstract int OpenGLListLength{get;}

      /// <summary>
      /// Free the geometry structure memory
      /// </summary>
      public abstract void FreeGeometry();

		/// <summary>
		/// 3D Object Info
		/// </summary>
		public ObjectInfo ObjectInfo
		{
			get
			{
				return m_objectInfo;
			}
			set
			{
				if (value != null)
				{
					m_objectInfo = value;
				}
			}
		}

      /// <summary>
      /// Calculates the normal vector of the specified triangle
      /// </summary>
      /// <param name="v1">vector 1</param>
      /// <param name="v2">vector 2</param>
      /// <param name="v3">vector 2</param>
      /// <returns>The normal vector</returns>
      public static Point TriangleNormal(Point v1, Point v2, Point v3)
      {
         // triangle normal calc
         float v1x = v3.X - v1.X;
         float v1y = v3.Y - v1.Y;
         float v1z = v3.Z - v1.Z;
         float v2x = v2.X - v1.X;
         float v2y = v2.Y - v1.Y;
         float v2z = v2.Z - v1.Z;
         
         // cross product v1->v2
         Point cp = new Point(
            (v1y * v2z) - (v1z * v2y),
            (v1z * v2x) - (v1x * v2z),
            (v1x * v2y) - (v1y * v2x));
         
         // normalise vector
         //Normalize(cp);
         
         return cp;
      }

		/// <summary>
		/// Find the mid-point of the specified polygon (MUST be 3 or 4 points only)
		/// </summary>
		/// <param name="poly">Polygon</param>
		/// <param name="points">Points[]</param>
		/// <returns>Mid-point</returns>
		public static Point PolygonMidPoint(Polygon poly, Point[] points)
		{
			Debug.Assert(poly != null && (poly.VertexCount == 3 || poly.VertexCount == 4), "Polygon not valid tri or quad!");
			if (poly.VertexCount == 3)
			{
				Point p1 = points[poly.Vertices[0]];
				Point p2 = points[poly.Vertices[1]];
				Point p3 = points[poly.Vertices[2]];
				return new Point( (p1.X + p2.X + p3.X) * 0.33333f,
										(p1.Y + p2.Y + p3.Y) * 0.33333f,
										(p1.Z + p2.Z + p3.Z) * 0.33333f );
			}
			else
			{
				Point p1 = points[poly.Vertices[0]];
				Point p2 = points[poly.Vertices[1]];
				Point p3 = points[poly.Vertices[2]];
				Point p4 = points[poly.Vertices[3]];
				return new Point( (p1.X + p2.X + p3.X + p4.X) * 0.25f,
										(p1.Y + p2.Y + p3.Y + p4.Y) * 0.25f,
										(p1.Z + p2.Z + p3.Z + p4.Z) * 0.25f );
			}
		}

		/// <summary>
		/// Return whether the specified polys are adjacent
		/// They are if they share at least two common points
		/// </summary>
		/// <param name="poly1">Polygon</param>
		/// <param name="poly2">Polygon</param>
		/// <returns>If adjacent</returns>
		public static bool IsAdjacentToo(Polygon poly1, Polygon poly2)
		{
			int commonPoints = 0;
			for (int vertexIndex=0; vertexIndex<poly1.VertexCount; vertexIndex++)
			{
				int poly1Vertex = poly1.Vertices[vertexIndex];
				for (int innerVertexIndex=0; innerVertexIndex<poly2.VertexCount; innerVertexIndex++)
				{
					if (poly1Vertex == poly2.Vertices[innerVertexIndex])
					{
						commonPoints++;
						break;
					}
				}
				if (commonPoints >= 2) break;
			}
			return (commonPoints >= 2);
		}

		/// <summary>
		/// Return the mid-point for two vertices of an edge
		/// </summary>
		/// <param name="p1">Point</param>
		/// <param name="p2">Point</param>
		/// <returns>Mid-point value</returns>
		public static Point EdgeMidPoint(Point p1, Point p2)
		{
			return new Point( (p1.X + p2.X) * 0.5f,
									(p1.Y + p2.Y) * 0.5f,
									(p1.Z + p2.Z) * 0.5f );
		}

		/// <summary>
		/// Averages point values
		/// </summary>
		/// <param name="p1">Point</param>
		/// <param name="p2">Point</param>
		/// <param name="p3">Point</param>
		/// <param name="p4">Point</param>
		/// <returns>Averaged Point</returns>
		public static Point AveragePoint(Point p1, Point p2, Point p3, Point p4)
		{
			return new Point( (p1.X + p2.X + p3.X + p4.X) * 0.25f,
										 (p1.Y + p2.Y + p3.Y + p4.Y) * 0.25f,
										 (p1.Z + p2.Z + p3.Z + p4.Z) * 0.25f );
		}

		public static Point AveragePoint(Point p1, Point p2, Point p3)
		{
			return new Point( (p1.X + p2.X + p3.X) * 0.333333f,
										 (p1.Y + p2.Y + p3.Y) * 0.333333f,
										 (p1.Z + p2.Z + p3.Z) * 0.333333f );
		}

		public static Point AveragePoint(Point p1, Point p2)
		{
			return new Point( (p1.X + p2.X) * 0.5f,
										 (p1.Y + p2.Y) * 0.5f,
										 (p1.Z + p2.Z) * 0.5f );
		}
      
      /// <summary>
      /// Normalizes the specified vector
      /// </summary>
      /// <param name="v">vector to normalize</param>
      public static Point Normalize(Point v)
      {
         float d = (float)Math.Sqrt( (double)(v.X*v.X + v.Y*v.Y + v.Z*v.Z) );
         
         if (d != 0.0f && d != 1.0f)
         {
            v.X /= d;
            v.Y /= d;
            v.Z /= d;
         }
			
			return v;
         /*else if (d == 0.0f)
         {
            if (K3DGLForm.DEBUG) Trace.WriteLine("Null vector in normalization!");
         }*/
      }

      /// <summary>
      /// Return the dot product of two vectors
      /// </summary>
      /// <param name="u"></param>
      /// <param name="v"></param>
      /// <returns>float dot product</returns>
      public static float DotProduct(Point u, Point v)
      {
         return u.X*v.X + u.Y*v.Y + u.Z*v.Z;
      }

      /// <summary>
      /// Calculates if an >4 sided polygon is concave or not
      /// </summary>
      /// <param name="polygon">Polygon to check</param>
      /// <param name="points">Points referenced by polygons</param>
      /// <returns>True if concave, false if convex</returns>
      public static bool IsConcave(Polygon polygon, Point[] points)
      {
	      // work round the points in the polygon
	      // take 3 at a time and work out the sign
	      // if the sign of any of the other groups of three points is different
	      // then we have a concave polygon
	      // else its convex

	      float a1, b1, c1, a2, b2, c2;
	      float val;
	      int lastSign = 0;
	      int currentSign;
	      bool firstVertex = true;

	      // test rest of polygon 3 points groups against sign
	      for (int vertCount=0; vertCount<polygon.VertexCount-2; vertCount++)
	      {
		      // a1 = x1 - x2
            a1 = points[ polygon.Vertices[vertCount] ].X - points[ polygon.Vertices[vertCount + 1] ].X;

		      // b1 = y1 - y2
            b1 = points[ polygon.Vertices[vertCount] ].Y - points[ polygon.Vertices[vertCount + 1] ].Y;

		      // c1 = z1 - z2
            c1 = points[ polygon.Vertices[vertCount] ].Z - points[ polygon.Vertices[vertCount + 1] ].Z;

		      // a2 = x2 - x3
            a2 = points[ polygon.Vertices[vertCount + 1] ].X - points[ polygon.Vertices[vertCount + 2] ].X;

		      // b2 = y2 - y3
            b2 = points[ polygon.Vertices[vertCount + 1] ].Y - points[ polygon.Vertices[vertCount + 2] ].Y;

		      // c2 = z2 - z3
            c2 = points[ polygon.Vertices[vertCount + 1] ].Z - points[ polygon.Vertices[vertCount + 2] ].Z;

		      val = ((b1 * c2) - (c1 * b2)) + ((c1 * a2) - (a1 * c2)) + ((a1 * b2) - (b1 * a2));

		      if (firstVertex)
		      {
			      firstVertex = false;
               if (val > 0.0f)
               {
                  lastSign = 1;
               }
               else
               {
                  lastSign = -1;
               }
		      }
		      else
		      {
               if (val > 0.0f)
               {
                  currentSign = 1;
               }
               else
               {
                  currentSign = -1;
               }
               
               if (currentSign != lastSign)
               {
                  // get out immediately the sign changes
                  return true;
               }
		      }
	      }
      	
	      return false;
      }

		/// <summary>
		/// Create and populate the AdjacentPolygons collection for a list of polys
		/// </summary>
		/// <param name="polygons">ArrayList of Polygons to build for</param>
		public static void PopulateAdjacentPolygons(ArrayList polygons)
		{
			/*int totalPolys = polygons.Count;
			for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
			{
				Polygon polygon = (Polygon)polygons[polyIndex];
				if (polygon.AdjacentPolygons == null)
				{
					// create list for first time
					polygon.AdjacentPolygons = new AdjacentPolyCollection(polygon);
				}
				
				for (int innerPolyIndex=(polyIndex+1); innerPolyIndex<totalPolys; innerPolyIndex++)
				{
					if (polygon.AdjacentPolygons.Count == polygon.VertexCount)
					{
						// get out if we have max adjacent polys for this polygon
						break;
					}
					
					// if we set value to both polys here, can skip some polys in loop
					Polygon innerPoly = (Polygon)polygons[innerPolyIndex];
					if (IsAdjacentToo(polygon, innerPoly) == true)
					{
						polygon.AdjacentPolygons.AddAdjacentIndex(innerPolyIndex);
						if (innerPoly.AdjacentPolygons == null)
						{
							// create list for first time
							innerPoly.AdjacentPolygons = new AdjacentPolyCollection(innerPoly);
						}
						innerPoly.AdjacentPolygons.AddAdjacentIndex(polyIndex);
					}
				}
			}*/
		}

		/// <summary>
		/// Create and populate an EdgeMap collection for a list of polys
		/// </summary>
		/// <param name="polygons">ArrayList of Polygons to build for</param>
		/// <param name="edgeMap">The EdgeMap to use</param>
		public static void PopulatePolygonEdgeMap(ArrayList polygons, EdgeMap edgeMap)
		{
			int totalPolys = polygons.Count;
			for (int polyIndex=0; polyIndex<totalPolys; polyIndex++)
			{
				Polygon polygon = (Polygon)polygons[polyIndex];
				
				// walk and store the edges for this polygon
				int vertexCount = polygon.VertexCount;
				for (int vertexIndex=0; vertexIndex<vertexCount; vertexIndex++)
				{
					// TODO: try moving these inner stack variables out of the loops!
					int edgeVertex1 = polygon.Vertices[vertexIndex];
					int edgeVertex2;
					if (vertexIndex < vertexCount - 1)
					{
						edgeVertex2 = polygon.Vertices[vertexIndex + 1];
					}
					else
					{
						edgeVertex2 = polygon.Vertices[0];
					}
					
					edgeMap.AddPolyForEdge(edgeVertex1, edgeVertex2, polyIndex);
				}
			}
		}


      // ------------------------------------------------------------------------------
      // Private data

		private ObjectInfo m_objectInfo = new ObjectInfo();
   }
}
